/*
QuickAsm- An 8-bit cross-assembler

Copyright (C) 1997 Scott Wilson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "asm.h"

/* Global variables */
opcodeType	opcodes[1000];
int			opcodeCnt = 0;

int			PC = 0;
int			printLineNumbers = 0; 

labelType		*labelList = NULL;
expressionType	*expList = NULL;
lineType		*lineList = NULL;
lineType		*lastLine = NULL;

void placeValues(void)
{
	lineType	*theLine;
	char		*d1 = NULL, *d2 = NULL;
	int			hexNumber;
	char		temp[12];
	int			cnt;
	int			relPC;
	int			offset;
	
	theLine = lineList;
	while(theLine)
		{
		if(theLine->opcodes)
			{
			d1 = (char *) strstr(theLine->opcodes, "XX");
			if(d1)		/* has values to be placed */
				{
				if(strlen(d1) > 2)
					d2 = (char *) strstr(d1 + 2, "XX");
				else
					d2 = NULL;

				if(theLine->exp1)
					if(!theLine->exp1->value[0])
						{
						crunchExp(theLine->exp1->expression, theLine->exp1->value);
						if(!theLine->exp1->value[0])
							theLine->error = kExpressionErr;
						}

				if(theLine->exp2)
					if(!theLine->exp2->value[0])
						{
						crunchExp(theLine->exp2->expression, theLine->exp2->value);
						if(!theLine->exp2->value[0])
							theLine->error = kExpressionErr;
						}

				switch(theLine->expType)
					{
					case 'A':	/* 16 bit data */
					case 'J':
						sscanf(theLine->exp1->value, "%x", &hexNumber);
						sprintf(temp, "%.4X", hexNumber);
						d1[0] = temp[2];
						d1[1] = temp[3];
						d1[2] = temp[0];
						d1[3] = temp[1];
						break;
					case 'D':	/* 8 bit data */
						sscanf(theLine->exp1->value, "%x", &hexNumber);
						sprintf(temp, "%.2X", hexNumber);
						d1[0] = temp[0];
						d1[1] = temp[1];
						if(d2)
							{
							sscanf(theLine->exp2->value, "%x", &hexNumber);
							sprintf(temp, "%.2X", hexNumber);
							d2[0] = temp[0];
							d2[1] = temp[1];
							}
						break;
					case 'R':	/* 8 bit relative */
						relPC = ((lineType *)theLine->next)->address;
						sscanf(theLine->exp1->value, "%x", &hexNumber);
						offset = (hexNumber - relPC) & 0x00FF;
						sprintf(temp, "%.2X", offset);
						d1[0] = temp[0];
						d1[1] = temp[1];
						break;
					}
				}
			}
		theLine = theLine->next;
		}
}

char *convertBytes(char *inString)
{
	char	temp[255] = "";
	char	exp[32] = "";
	char	byte[32] = "";
	int		strCnt = 0;
	
	while(inString[strCnt] != 0)
		{
		if(inString[strCnt]!=',')
			{
			if( !isspace(inString[strCnt]) )
				strncat(exp, inString + strCnt, 1);
			}
		else{
			crunchExp(exp, byte);
			strcat(temp, byte);
			exp[0] = 0;
			}
		strCnt++;
		}
	if(exp[0])
		{
		crunchExp(exp, byte);
		strcat(temp, byte);
		exp[0] = 0;
		}
	return (char *) strdup(temp);
}

char *convertWords(char *inString)
{
	char	temp[255] = "";
	char	exp[32] = "";
	char	byte[32] = "";
	int		strCnt = 0;
	
	while(inString[strCnt] != 0)
		{
		if(inString[strCnt]!=',')
			{
			if( !isspace(inString[strCnt]) )
				strncat(exp, inString + strCnt, 1);
			}
		else{
			crunchExp(exp, byte);
			strcat(temp, byte);
			exp[0] = 0;
			}
		strCnt++;
		}
	if(exp[0])
		{
		crunchExp(exp, byte);
		if(strlen(byte) < 4)
			strcat(temp,"00");
		strcat(temp, byte);
		exp[0] = 0;
		}
	return (char *) strdup(temp);
}

char *convertString(char *inString)
{
	char	*out;
	int		num, cnt;
	char	temp[4];
	
	if( (inString == 0) || (inString[0] == 0) )
		return 0;
	out = (char *) malloc(strlen(inString)*2);
	out[0] = 0;
	for(cnt = 1; cnt < (strlen(inString) - 1); cnt++)
		{
		num = inString[cnt];
		sprintf(temp,"%.2X", num);
		strncat(out, temp, 2);
		}
	return out;
}

void crunchExp(char *exp, char *ans)
{
	char		value[32] = "", temp[32] = "";
	int			cnt = 0;
	int			hexValue;
	labelType	*theLabel;
	
	ans[0] = 0;
	if(isdigit(exp[0]) || (exp[0] == '$'))
		{
		if(exp[0] == '$')
			{
			exp[0] = '0';
			strncat(exp, "H", 1);
			}
		while( (cnt < strlen(exp)) && isalnum(exp[cnt]) )
			{
			strncat(value, exp + cnt, 1);
			cnt++;
			}
		switch(value[cnt - 1])
			{
				case 'H':
					value[cnt - 1] = 0;
					sscanf(value,"%x", &hexValue);
					break;
				case 'O':
					value[cnt - 1] = 0;
					sscanf(value,"%o", &hexValue);
					break;
				case 'D':
					value[cnt - 1] = 0;
				default:
					sscanf(value,"%d", &hexValue);
					break;
			}
		sprintf(temp, "%X", hexValue);
		strcpy(ans, temp);
		}
	else{	/* Right now if the expression doesn't start with a digit, then it is assumed to be a label */
		theLabel = labelList;
		while(theLabel)
			{
			if(strcmp(exp, theLabel->label) == 0)
				{
				strcpy(ans, theLabel->value);
				return ;
				}
			theLabel = theLabel->next;
			}
		}
}

void getExpressions(int index, char *operators, char *exp1, char *exp2)
{
	int		cnt1, cnt2 = 0;
	char	*string1, *string2;
	int		expCnt = 1;
	
	exp1[0] = 0;
	exp2[0] = 0;
	string2 = operators;
	string1 = opcodes[index].operators;
	
	if(index == 228)
		index = index;
	for(cnt1 = 0; cnt1 < strlen(string1); cnt1++)
		{
		if(string1[cnt1] != string2[cnt2])
			{
			if(string1[cnt1] == '~')
				{
				cnt1++;
				while( (cnt2 <= strlen(string2)) && (string1[cnt1] != string2[cnt2]) )
					{
					if(expCnt == 1)
						{	/* exp1 */
						strncat(exp1, string2 + cnt2, 1);
						}
					else{	/* exp2 */
						strncat(exp2, string2 + cnt2, 1);
						}
					cnt2++;
					}
				expCnt++;
				}
			}
		cnt2++;
		}
}

int findOpcode(char *mnemonic, char *operators)
{
	int		cnt;
	
	for(cnt = 0; cnt < opcodeCnt; cnt++)
		{
		while( ( strcmp(mnemonic, opcodes[cnt].mnemonic) == 0 ) && (cnt < opcodeCnt) )
			{
			if( compareString(opcodes[cnt].operators, operators) == 0 )
				return cnt;
			cnt++;
			}
		}
	return -1;
}

/*	Compares two strings, using '~' as a wild character 		*/
/*	NOTE: '~' is only wild if it is contained within string1 	*/
int	 compareString(char *string1, char *string2)
{
	int		cnt1, cnt2 = 0;
	
	if(strcmp(string1, "(~),HL") == 0)
		{
		cnt1 = 0;
		}
	
	if( (strcmp(string1, string2) == 0) )
		return 0;
	if( (string2[0] == 0) && (string1[0] == 0) )
		return 0;
	for(cnt1 = 0; cnt1 < strlen(string1); cnt1++)
		{
		if(string1[cnt1] != string2[cnt2])
			{
			if(string1[cnt1] == '~')
				{
				cnt1++;
				while( (cnt2 <= strlen(string2)) && (string1[cnt1] != string2[++cnt2]) );
				if(string1[cnt1] != string2[cnt2])
					return strcmp(string1, string2);
				if(string1[cnt1] == 0)
					return 0;
				}
			else{		/* Not Equal */
				return 1;
				}
			}
		cnt2++;
		}
	if(string1[cnt1] == string2[cnt2])
		return 0;
	else
		return 1;
}

labelType *newLabel(char *label, char *value)
{
	labelType	*theLabel;
	
	if(label && label[0])
		{
		theLabel = (labelType *) malloc(sizeof(labelType));
		theLabel->next = labelList;
		labelList = theLabel;

		theLabel->label = (char *) strdup(label);
		strcpy(theLabel->value, value);
		return(theLabel);
		}
	else return NULL;
}


expressionType *newExpression(char *exp, char *value)
{
	expressionType	*theExp;
	
	if(exp && exp[0])
		{
		theExp = (expressionType *) malloc(sizeof(expressionType));
		theExp->next = expList;
		expList = theExp;

		if(value[0] = 0)
			theExp->expression = NULL;
		else
			theExp->expression = (char *) strdup(exp);
		strcpy(theExp->value, value);
		return(theExp);
		}
	else return NULL;
}
void processLine(char *line)
{
	lineType		*theLine;
	labelType		*theLabel;
	expressionType	*theExp;
	char			mnemonic[12] = "";
	char			operators[255] = "";
	char			exp1[255] = "", exp2[255] = "";
	char			label[32] = "";
	char			temp[12] = "";
	int				index, strCnt, test;
	
	/* Linked list stuff */
	theLine = (lineType *) malloc( sizeof(lineType) );
	theLine->next = NULL;
	if(lineList == NULL)
		{
		lineList = theLine;
		lastLine = theLine;
		}
	else{
		lastLine->next = theLine;
		lastLine = theLine;
		}
	
	/* initialize the new line record */
	theLine->address = PC;
	theLine->opcodes = NULL;
	theLine->line = (char *) strdup(line);
	theLine->exp1 = NULL;
	theLine->exp2 = NULL;
	theLine->label = NULL;
	theLine->error = 0;
	theLine->print = 1;
	for(strCnt = 0; strCnt < strlen(line); strCnt++)
		{
		if( (line[strCnt] == ';') || (line[strCnt] == '"') )
			strCnt = strlen(line);
		else
			line[strCnt] = toupper(line[strCnt]);
		}
	strCnt = 0;
	
	if( isalpha(line[strCnt]) )		/* This line is labeled */
		{
		while( isalnum(line[strCnt]) )
			strncat(label, line + strCnt++, 1);
		if(line[strCnt] == ':')
			strCnt++;
		}
		
	/* Advance through white space */
	while( (line[strCnt] == ' ') || (line[strCnt] == '\t') && (line[strCnt] != ';') ) strCnt++; 
	
	/* Read in mnemonic */
	while( (line[strCnt] != ' ') && (line[strCnt] != '\t') && (line[strCnt] != ';') && (strCnt < strlen(line)) )
		{
		strncat(mnemonic, line + strCnt, 1);
		strCnt++;
		}
	
	/* Read in operators, ignore all white space */
	while( (line[strCnt] != ';') && (strCnt < strlen(line)) )
		{
		while( ((line[strCnt] == ' ') || (line[strCnt] == '\t')) && (strCnt < strlen(line)) && (line[strCnt] != ';') )
			strCnt++;
		if(line[strCnt] != ';')
			strncat(operators, line + strCnt, 1);
		else
			strCnt = strlen(line);
		strCnt++;
		}

	if(mnemonic[0])
		{
		index = findOpcode(mnemonic, operators);
		if(index > -1)
			{
			sprintf(temp,"%X",PC);
			theLine->label = newLabel(label, temp);
			theLine->opcodes = (char *) strdup(opcodes[index].byteCode);
			PC += strlen(theLine->opcodes) / 2;
			getExpressions(index, operators, exp1, exp2);
			crunchExp(exp1, temp);
			theLine->exp1 = newExpression(exp1, temp);
			crunchExp(exp2, temp);
			theLine->exp2 = newExpression(exp2, temp);
			theLine->expType = opcodes[index].type;
			}
		else{
			if( (strcmp(mnemonic, ".ORG") == 0) ||  (strcmp(mnemonic, "ORG") == 0))
				{
				crunchExp(operators, temp);
				if(temp)
					{
					sscanf(temp,"%x", &PC);
					theLine->address = PC;

					theLine->label = newLabel(label, temp);
					}
				else{
					theLine->error = kExpressionErr;
					}
				}
			else if( (strcmp(mnemonic, ".EQU") == 0)  ||  (strcmp(mnemonic, "EQU") == 0))
				{
				crunchExp(operators, temp);
				theLine->label = newLabel(label, temp);
				}
			else if( (strcmp(mnemonic, ".DB") == 0)  ||  (strcmp(mnemonic, "DB") == 0))
				{
				if(operators[0] == '"')
					{			/* A string is defined */
					theLine->opcodes = convertString(operators);
					PC += strlen(theLine->opcodes) / 2;
					}
				else{			/* the data is byte[s] seperated by commas */
					theLine->opcodes = convertBytes(operators);
					PC += strlen(theLine->opcodes) / 2;
					}
				sprintf(temp,"%X",PC);
				theLine->label = newLabel(label, temp);
				}
			else if( (strcmp(mnemonic, ".DW") == 0)  ||  (strcmp(mnemonic, "DW") == 0))
				{
				theLine->opcodes = convertWords(operators);
				PC += strlen(theLine->opcodes) / 2;
				sprintf(temp,"%X",PC);
				theLine->label = newLabel(label, temp);
				}
			else if( (strcmp(mnemonic, ".DS") == 0)  ||  (strcmp(mnemonic, "DS") == 0))
				{
				crunchExp(operators, temp);
				sscanf(temp,"%x", &test);
				PC += test;
				theLine->address = PC;		
				
				sprintf(temp,"%X",PC);
				theLine->label = newLabel(label, temp);
				}
			else if( (strcmp(mnemonic, ".END") == 0)  ||  (strcmp(mnemonic, "END") == 0))
				{
				}
			else{	/* No such opcode */
				theLine->error = kNoOpcodeErr;
				}
			}
		}
	else{
		theLine->print = 0;
		}
}

int  myStringCmp(const void *s1, const void *s2)
{
	opcodeType	*op1 = (opcodeType *) s1;
	opcodeType	*op2 = (opcodeType *) s2;
	int			res;
	
	res = strcmp(op1->mnemonic, op2->mnemonic);
	if(res == 0)
		res = strcmp(op1->operators, op2->operators);
		if(res == 0)
			return strlen(op1->byteCode) - strlen(op2->byteCode);
	return res;
}