/*
QuickAsm- An 8-bit cross-assembler

Copyright (C) 1997 Scott Wilson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>

#define	kExpressionErr		1
#define	kNoOpcodeErr		2

typedef	struct{
			char	mnemonic[12];
			char	operators[12];
			char	byteCode[12];
			char	type;
		}opcodeType;
		
typedef	struct{
			char	*label;			/* Name of label */
			char	value[8];		/* Represented as ASCII (eg 0xf8 = "f8") */
			void	*next;
		}labelType;			
		
typedef	struct{
			char	*expression;
			char	value[8];
			void	*next;
		}expressionType;
		
typedef	struct{
			short			address;			/* Stored in binary form */
			char			*opcodes;			/* Stored in ASCII form */
			char			*line;				/* the entire line */
			char			expType;
			expressionType	*exp1;
			expressionType	*exp2;
			labelType		*label;
			int				error;
			int				print;
			void			*next;
		}lineType;


/* Global variables */
extern	opcodeType	opcodes[1000];
extern	int			opcodeCnt;

extern	int			PC;
extern	int			printLineNumbers; 

extern	labelType		*labelList;
extern	expressionType	*expList;
extern	lineType		*lineList;
extern	lineType		*lastLine;

/* Procedure declarations */
void processLine(char *line);
int  findOpcode(char *mnemonic, char *operators);
int	 compareString(char *string1, char *string2);
void crunchExp(char *exp, char *ans);
void getExpressions(int index, char *operators, char *exp1, char *exp2);
void placeValues(void);
int  myStringCmp(const void *s1, const void *s2);
char *convertString(char *inString);
char *convertBytes(char *inString);
char *convertWords(char *inString);

labelType 		*newLabel(char *label, char *value);
expressionType	*newExpression(char *exp, char *value);

