/*
QuickAsm- An 8-bit cross-assembler

Copyright (C) 1997 Scott Wilson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "unix.h"

#define	kHexBytes	0x18

void makeHexLine(char *line, char *data, int addr);

void main (int argc, char *argv[])
{
	char	hex[255], list[255], in[255] = "boot.asm";
	
	if(argc >= 2)
		strcpy(in, argv[1]);
	
	if(argc == 1)
		argc = 2;
		
	if(argc >= 2)
		{
		readBrainFile(kBrainFile);	

		assembleFile(in);

		placeValues();

		strcpy(hex, in);
		if(strcmp(hex + strlen(hex) - 4, ".asm") == 0)
			hex[strlen(hex) - 4] = 0;
		strcat(hex, ".hex");
		makeHexFile(hex);

		strcpy(list, in);
		if(strcmp(list + strlen(list) - 4, ".asm") == 0)
			list[strlen(list) - 4] = 0;
		strcat(list, ".lst");
		printFile(list);
		}
	else{
		printf(" Usage:\n");
		printf("    asm infile.asm\n");
		}
}

void makeHexLine(char *line, char *data, int addr)
{
	char	temp[16];
	int		size;
	int		cksum = 0;
	int		byte = 0;
	int		offset = 0;
	
	line[0] = 0;
	strcat(line, ":");
	size = strlen(data) / 2;
	sprintf(temp,"%.2X", size);
	strcat(line, temp);
	sprintf(temp,"%.4X",addr);
	strcat(line, temp);
	strcat(line,"00");
	strcat(line, data);
	for(offset = 1; offset < strlen(line); offset += 2)
		{
		strncpy(temp, line + offset, 2);
		temp[2] = 0;
		sscanf(temp, "%x", &byte);
		cksum += byte;
		}
	/* I know that there is a 2's compliment function in C, but I don't know what it is... */
	cksum = ~cksum;
	cksum++;
	cksum = cksum & 0x00FF;
	
	sprintf(temp,"%.2X", cksum);
	strcat(line, temp);
}

void makeHexFile(char *hexFile)
{
	char		data[(kHexBytes * 2) + 2] = "";
	int			bytes;
	int			offset = 0;
	char		line[80] = "";
	int			curPC = 0, startPC = 0;
	lineType	*theLine;
	FILE		*out;
	
	theLine = lineList;
	out = fopen(hexFile, "w");
	while(theLine)
		{
		if(theLine->opcodes)
			{
			if(curPC != (theLine->address + (offset / 2)) )
				{
				makeHexLine(line, data, startPC);
				startPC = theLine->address;
				curPC = startPC;
				data[0] = 0;
				if(line[0])
					fprintf(out,"%s\n", line);
				}
			while( (offset < strlen(theLine->opcodes)) && ( (kHexBytes * 2) > strlen(data)) )
				{
				strncat(data, theLine->opcodes + offset, 2);
				offset += 2;
				curPC++;
				}
			if( strlen(data) >= (kHexBytes * 2) )
				{
				makeHexLine(line, data, startPC);
				startPC = curPC;
				data[0] = 0;
				if(line[0])
					fprintf(out,"%s\n", line);
				}
			if(offset >= strlen(theLine->opcodes))
				{
				theLine = theLine->next;
				offset = 0;
				}
			}
		else
			theLine = theLine->next;
		}
	if(data[0])
		{
		makeHexLine(line, data, startPC);
		startPC = curPC;
		data[0] = 0;
		if(line[0])
			fprintf(out,"%s\n", line);
		}
	fprintf(out,":00000001FF");
	fclose(out);
}

void printFile(char *listFile)
{
	lineType	*theLine;
	FILE		*out;
	int			cnt;
	int			lineCnt = 1;
	int			spaceCnt;
	
	theLine = lineList;
	out = fopen(listFile, "w");
	while(theLine)
		{
		if(theLine->error)
			{
			fprintf(out,"\nError: %d\n  ", theLine->error);
			printf("\nError: %d\n  ", theLine->error);
			printf("%s\n", theLine->line);
			}
		if(printLineNumbers)
			fprintf(out,"%.4d   ", lineCnt++);
		if(theLine->print)
			{
			fprintf(out, "%.4X ", theLine->address);
			if(theLine->opcodes)
				{
				for(cnt = 0; (cnt < (strlen(theLine->opcodes) / 2)) && (cnt < 4); cnt++)
					{
					fprintf(out, "%.2s", theLine->opcodes + (cnt * 2));
					if( (cnt*2 + 2) < strlen(theLine->opcodes) )
						fprintf(out, " ");
					}
				spaceCnt = 11 - (strlen(theLine->opcodes) / 2) - strlen(theLine->opcodes);
				for(cnt = 0; cnt < spaceCnt; cnt++)
					fprintf(out, " ");
				if( strlen(theLine->opcodes) <= 8)
					fprintf(out, " ");
				if( (strlen(theLine->opcodes) > 0) && (strlen(theLine->opcodes) < 8))
					fprintf(out, " ");
				}
			else
				for(cnt = 0; cnt < 12; cnt++)
					fprintf(out, " ");
			}
		else{
			for(cnt = 0; cnt < 17; cnt++)
				fprintf(out, " ");
			}
		fprintf(out, "%s\n", theLine->line);
		if(theLine->error)
			fprintf(out,"\n");
		theLine = theLine->next;
		}
	fclose(out);
}

void assembleFile(char *asmFile)
{
	FILE	*in;
	char	inStr[256];
	
	in = fopen(asmFile,"r");
	while( fgets(inStr, 256, in) != 0 )
		{
		inStr[strlen(inStr)-1] = 0;
		processLine(inStr);
		}
	fclose(in);
}

void readBrainFile(char *brainFile)
{
	FILE	*in;
	char	inStr[128];
	int		strCnt;
	
	opcodeCnt = -1;
	in = fopen(brainFile, "r");
	while( fgets(inStr, 128, in) != 0 )
		{
		opcodes[opcodeCnt + 1].mnemonic[0] = 0;
		opcodes[opcodeCnt + 1].operators[0] = 0;
		opcodes[opcodeCnt + 1].byteCode[0] = 0;
		opcodes[opcodeCnt + 1].type = 0;
		strCnt = 0;
		
		while( ((inStr[strCnt] == '\t') || (inStr[strCnt] == ' ')) && (inStr[strCnt] != 0) && (inStr[strCnt] != ';') && (inStr[strCnt] != '\n') ) strCnt++;
		if(inStr[strCnt] == ';')
			strCnt = strlen(inStr);
		else
			opcodeCnt++;
			
		while( (inStr[strCnt] != 0) && (inStr[strCnt] != '\t')  && (inStr[strCnt] != ' ') && (inStr[strCnt] != '\n') && (inStr[strCnt] != ';'))
			{
			strncat(opcodes[opcodeCnt].byteCode, inStr + strCnt, 1);
			strCnt++;
			}
			
		while( ((inStr[strCnt] == '\t') || (inStr[strCnt] == ' ')) && (inStr[strCnt] != 0) && (inStr[strCnt] != ';') && (inStr[strCnt] != '\n') ) strCnt++;
		if(inStr[strCnt] == ';')
			strCnt = strlen(inStr);
			
		while( (inStr[strCnt] != 0) && (inStr[strCnt] != '\t')  && (inStr[strCnt] != ' ') && (inStr[strCnt] != '\n') && (inStr[strCnt] != ';'))
			{
			strncat(opcodes[opcodeCnt].mnemonic, inStr + strCnt, 1);
			strCnt++;
			}
			
		while( ((inStr[strCnt] == '\t') || (inStr[strCnt] == ' ')) && (inStr[strCnt] != 0) && (inStr[strCnt] != ';') && (inStr[strCnt] != '\n') ) strCnt++;
		if(inStr[strCnt] == ';')
			strCnt = strlen(inStr);
			
		while( (inStr[strCnt] != 0) && (inStr[strCnt] != '\t')  && (inStr[strCnt] != ' ') && (inStr[strCnt] != '\n') && (inStr[strCnt] != ';'))
			{
			if(inStr[strCnt] == '%')
				{
				opcodes[opcodeCnt].type = inStr[++strCnt];
				strncat(opcodes[opcodeCnt].operators, "~", 1);
				}
			else
				strncat(opcodes[opcodeCnt].operators, inStr + strCnt, 1);
			strCnt++;
			}
		}
	fclose(in);
	qsort(opcodes, opcodeCnt, sizeof(opcodeType), myStringCmp);
}
